// ======================================================================
//
// Xform.cpp
// jeff grills
//
// copyright 1998 origin systems inc
//
// ======================================================================

#include "First.h"
#include "Xform.h"

#include "Gl.def"
#include "Quaternion.h"

// ======================================================================

#define USE_INLINE_ASM 1

// ======================================================================

const Xform Xform::identity;

// ======================================================================
// Move relative to the current frame of reference
//
// Remarks:
//
//   Transforms the specified translation from the frame-of-reference of this transform into parent space,
//   and then adds that result to the position of this transform.
//
// See also:
//
//   moveAbsolute(), moveRelative()

void Xform::move(
	const Vector &vec // translation in the transform's space
	)
{
	Vector v;

	rotate(vec, &v);
	moveRelative(v);
}

// ----------------------------------------------------------------------
// Move relative to the current frame of reference
//
// Remarks:
//
//   Transforms the specified translation from the frame-of-reference of this transform into parent space,
//   and then adds that result to the position of this transform.
//
// See also:
//
//   moveAbsolute(), moveRelative()

void Xform::move(
	real dx,  // x translation in the transform's space
	real dy,  // y translation in the transform's space
	real dz   // z translation in the transform's space
	)
{
	Vector v;

	rotate(Vector(dx, dy, dz), &v);
	moveRelative(v);
}

// ----------------------------------------------------------------------
// Rotate this transform around the X axis
//
// Remarks:
//
//   This routine is highly optimized, especially for the Win95 platform.

void Xform::pitch(
	real radians  // amount to rotate in radians
	)
{
	// 1  0  0  0
	// 0  C -S  0
	// 0  S  C  0
	// 0  0  0  1

	// here we do a custom optimized matrix multiplication against this matrix

#if PLATFORM == PLATFORM_WIN95

	// lint can't see the variable referenced in the inline assembly
	UNREF(radians);

	__asm
	{
		fld radians                       // rads
		fsincos                           // cos  sin
		mov eax, this
		fld st(1)                         // sin  cos  sin
		fchs                              // -sin  cos  sin

		fld DWORD PTR [eax+0+4]           // m01  -sin  cos  sin
		fmul st(0), st(2)                 // m01*cos  -sin  cos  sin
		fld DWORD PTR [eax+0+8]           // m02  m01*cos  -sin  cos  sin
		fmul st(0), st(4)                 // m02*sin  m01*cos  -sin  cos  sin
		fld DWORD PTR [eax+0+4]           // m01  m02*sin  m01*cos  -sin  cos  sin
		fmul st(0), st(3)                 // m01*-sin  m02*sin  m01*cos  -sin  cos  sin
		fxch st(2)                        // m01*cos  m02*sin  m01*-sin  -sin  cos  sin
		faddp st(1), st(0)                // m01*cos+m02*sin  m01*-sin  -sin  cos  sin
		fld DWORD PTR [eax+0+8]           // m02  m01*cos+m02*sin  m01*-sin  -sin  cos  sin
		fmul st(0), st(4)                 // m02*cos  m01*cos+m02*sin  m01*-sin  -sin  cos  sin
		fxch st(1)                        // m01*cos+m02*sin  m02*cos  m01*-sin  -sin  cos  sin
		fstp DWORD PTR [eax+0+4]          // m02*cos  m01*-sin  -sin  cos  sin
		faddp st(1), st(0)                // m02*cos+m01*-sin  -sin  cos  sin
		fld DWORD PTR [eax+16+4]          // m11  m02*cos+m01*-sin  -sin  cos  sin
		fmul st(0), st(3)                 // m11*cos  m02*cos+m01*-sin  -sin  cos  sin
		fxch st(1)                        // m02*cos+m01*-sin  m11*cos  -sin  cos  sin
		fstp DWORD PTR [eax+0+8]          // m11*cos  -sin  cos  sin
		fld DWORD PTR [eax+16+8]          // m12  m11*cos  -sin  cos  sin
		fmul st(0), st(4)                 // m12*sin  m11*cos  -sin  cos  sin
		fld DWORD PTR [eax+16+4]          // m11 	m12*sin  m11*cos  -sin  cos  sin
		fmul st(0), st(3)                 // m11*-sin  m12*sin  m11*cos  -sin  cos  sin
		fxch st(2)                        // m11*cos  m12*sin  m11*-sin  -sin  cos  sin
		faddp st(1), st(0)                // m11*cos+m12*sin  m11*-sin  -sin  cos  sin
		fld DWORD PTR [eax+16+8]          // m12  m11*cos+m12*sin  m11*-sin  -sin  cos  sin
		fmul st(0), st(4)                 // m12*cos  m11*cos+m12*sin  m11*-sin  -sin  cos  sin
		fxch st(1)                        // m11*cos+m12*sin  m12*cos  m11*-sin  -sin  cos  sin
		fstp DWORD PTR [eax+16+4]         // m12*cos  m11*-sin  -sin  cos  sin
		faddp st(1), st(0)                // m12*cos+m11*-sin  -sin  cos  sin
		fld DWORD PTR [eax+16+16+4]       // m21  m12*cos+m11*-sin  -sin  cos  sin
		fmul st(0), st(3)                 // m21*cos  m12*cos+m11*-sin  -sin  cos  sin
		fxch st(1)                        // m11*-sin+m12*cos  m21*cos   -sin  cos  sin
		fstp DWORD PTR [eax+16+8]         // m21*cos   -sin  cos  sin
		fld DWORD PTR [eax+16+16+8]       // m22  m21*cos   -sin  cos  sin
		fmul st(0), st(4)                 // m22*sin  m21*cos   -sin  cos  sin
		fld DWORD PTR [eax+16+16+4]       // m21  m22*sin  m21*cos   -sin  cos  sin
		fmul st(0), st(3)                 // m21*-sin  m22*sin  m21*cos   -sin  cos  sin
		fxch st(2)                        // m21*cos  m22*sin  m21*-sin  -sin  cos  sin
		faddp st(1), st(0)                // m21*cos+m22*sin  m21*-sin  -sin  cos  sin
		fld DWORD PTR [eax+16+16+8]       // m22  m21*cos+m22*sin  m21*-sin  -sin  cos  sin
		fmul st(0), st(4)                 // m22*cos  m21*cos+m22*sin  m21*-sin  -sin  cos  sin
		fxch st(1)                        // m21*cos+m22*sin  m22*cos  m21*-sin  -sin  cos  sin
		fstp DWORD PTR [eax+16+16+4]      // m22*cos  m21*-sin  -sin  cos  sin
		faddp st(1), st(0)                // m22*cos+m21*-sin  -sin  cos  sin
		fxch st(3)                        // sin  -sin  cos  m22*cos+m21*-sin  
		fstp st(0)                        // -sin  cos  m22*cos+m21*-sin  
		fstp st(0)                        // cos  m22*cos+m21*-sin  
		fstp st(0)                        // m22*cos+m21*-sin  
		fstp DWORD PTR [eax+16+16+8]      // empty
	}

#else

#error "xform::pitch NYF"

	real s, c, a, b;

	s = sin(radians);
	c = cos(radians);

#endif
}

// ----------------------------------------------------------------------
// Rotate this transform around the Y axis
//
// Remarks:
//
//   This routine is highly optimized, especially for the Win95 platform.

void Xform::yaw(
	real radians  // amount to rotate in radians
	)
{
	// this is the format of a yaw-only matrix:

	//  C  0  S  0
	//  0  1  0  0
	// -S  0  C  0
	//  0  0  0  1

	// here we do a custom optimized matrix multiplication against this matrix

#if PLATFORM == PLATFORM_WIN95

	// lint can't see the variable referenced in the inline assembly
	UNREF(radians);

	__asm
	{
		fld radians                       // rads
		fsincos                           // cos  sin
		mov eax, this
		fld st(1)                         // sin  cos  sin
		fchs                              // -sin  cos  sin

		fld DWORD PTR [eax+0+0]           // m00  -sin  cos  sin
		fmul st(0), st(2)                 // m00*cos  -sin  cos  sin
		fld DWORD PTR [eax+0+8]           // m02  m00*cos  -sin  cos  sin
		fmul st(0), st(2)                 // m02*-sin  m00*cos  -sin  cos  sin
		fld DWORD PTR [eax+0+0]           // m00  m02*-sin  m00*cos  -sin  cos  sin
		fmul st(0), st(5)                 // m00*sin  m02*-sin  m00*cos  -sin  cos  sin
		fxch st(2)                        // m00*cos  m02*-sin  m00*sin  -sin  cos  sin
		faddp st(1), st(0)                // m00*cos+m02*-sin  m00*sin  -sin  cos  sin
		fld DWORD PTR [eax+0+8]           // m02  m00*cos+m02*-sin  m00*sin  -sin  cos  sin
		fmul st(0), st(4)                 // m02*cos  m00*cos+m02*-sin  m00*sin  -sin  cos  sin
		fxch st(1)                        // m00*cos+m02*-sin  m02*cos  m00*sin  -sin  cos  sin
		fstp DWORD PTR [eax+0+0]          // m02*cos  m00*sin  -sin  cos  sin
		faddp st(1), st(0)                // m02*cos+m00*sin  -sin  cos  sin
		fld DWORD PTR [eax+16+0]          // m10  m02*cos+m00*sin  -sin  cos  sin
		fmul st(0), st(3)                 // m10*cos  m02*cos+m00*sin  -sin  cos  sin
		fxch st(1)                        // m02*cos+m00*sin  m10*cos  -sin  cos  sin
		fstp DWORD PTR [eax+0+8]          // m10*cos  -sin  cos  sin
		fld DWORD PTR[eax+16+8]           // m12  m10*cos  -sin  cos  sin
		fmul st(0), st(2)                 // m12*-sin  m10*cos  -sin  cos  sin
		fld DWORD PTR [eax+16+0]          // m10  m12*-sin  m10*cos  -sin  cos  sin
		fmul st(0), st(5)                 // m10*sin  m12*-sin  m10*cos  -sin  cos  sin
		fxch st(2)                        // m10*cos  m12*-sin  m10*sin  -sin  cos  sin
		faddp st(1), st(0)                // m10*cos+m12*-sin  m10*sin  -sin  cos  sin
		fld DWORD PTR [eax+16+8]          // m12  m10*cos+m12*-sin  m10*sin  -sin  cos  sin
		fmul st(0), st(4)                 // m12*cos  m10*cos+m12*-sin  m10*sin  -sin  cos  sin
		fxch st(1)                        // m10*cos+m12*-sin  m12*cos  m10*sin  -sin  cos  sin
		fstp DWORD PTR [eax+16+0]         // m12*cos  m10*sin  -sin  cos  sin
		faddp st(1), st(0)                // m12*cos+m10*sin  -sin  cos  sin
		fld DWORD PTR [eax+16+16+0]       // m20  m12*cos+m10*sin  -sin  cos  sin
		fmul st(0), st(3)                 // m20*cos  m12*cos+m10*sin  -sin  cos  sin
		fxch st(1)                        // m12*cos+m10*sin  m20*cos  -sin  cos  sin
		fstp DWORD PTR [eax+16+8]         // m20*cos  -sin  cos  sin
		fld DWORD PTR [eax+16+16+8]       // m22  m20*cos  -sin  cos  sin
		fmul st(0), st(2)                 // m22*-sin  m20*cos  -sin  cos  sin
		fld DWORD PTR [eax+16+16+0]       // m20  m22*-sin  m20*cos  -sin  cos  sin
		fmul st(0), st(5)                 // m20*sin  m22*-sin  m20*cos  -sin  cos  sin
		fxch st(2)                        // m20*cos  m22*-sin  m20*sin  -sin  cos  sin
		faddp st(1), st(0)                // m20*cos+m22*-sin  m20*sin  -sin  cos  sin
		fld DWORD PTR [eax+16+16+8]       // m22  m20*cos+m22*-sin  m20*sin  -sin  cos  sin
		fmul st(0), st(4)                 // m22*cos  m20*cos+m22*-sin  m20*sin  -sin  cos  sin
		fxch st(1)                        // m20*cos+m22*-sin  m22*cos  m20*sin  -sin  cos  sin
		fstp DWORD PTR [eax+16+16+0]      // m22*cos  m20*sin  -sin  cos  sin
		faddp st(1), st(0)                // m22*cos+m20*sin  -sin  cos  sin
		fxch st(3)                        // -sin  cos  sin  m22*cos+m20*sin
		fstp st(0)                        // cos  sin  m22*cos+m20*sin
		fstp st(0)                        // sin  m22*cos+m20*sin
		fstp st(0)                        // m22*cos+m20*sin
		fstp DWORD PTR [eax+16+16+8]      // empty
	}

#else

	real s, c, a, b;

	s = sin(radians);
	c = cos(radians);

	a = matrix[0][0];
	b = matrix[0][2];
	matrix[0][0] = a * c + b * -s;
	matrix[0][2] = a * s + b *  c;

	a = matrix[1][0];
	b = matrix[1][2];
	matrix[1][0] = a * c + b * -s;
	matrix[1][2] = a * s + b *  c;

	a = matrix[2][0];
	b = matrix[2][2];
	matrix[2][0] = a * c + b * -s;
	matrix[2][2] = a * s + b *  c;

#endif
}

// ----------------------------------------------------------------------
// Rotate this transform around the Z axis
//
// Remarks:
//
//   This routine is highly optimized, especially for the Win95 platform.

void Xform::roll(
	real radians  // amount to rotate in radians
	)
{
	// this is the format of a roll-only matrix:

	//  c -s  0  0
	//  s  c  0  0
	//  0  0  1  0
	//  0  0  0  1

	// here we do a custom optimized matrix multiplication against this matrix

#if PLATFORM == PLATFORM_WIN95

	// lint can't see the variable referenced in the inline assembly
	UNREF(radians);

	__asm
	{
		fld radians                       // rads
		fsincos                           // cos  sin
		mov eax, this
		fld st(1)                         // sin  cos  sin
		fchs                              // -sin  cos  sin

		fld DWORD PTR [eax+0+0]           // m00  -sin  cos  sin
		fmul st(0), st(2)                 // m00*cos  -sin  cos  sin
		fld DWORD PTR [eax+0+4]           // m01  m00*cos  -sin  cos  sin
		fmul st(0), st(4)                 // m01*sin  m00*cos  -sin  cos  sin
		fld DWORD PTR [eax+0+0]           // m00  m01*sin  m00*cos  -sin  cos  sin
		fmul st(0), st(3)                 // m00*-sin  m01*sin  m00*cos  -sin  cos  sin
		fxch st(2)                        // m00*cos  m01*sin  m00*-sin  -sin  cos  sin
		faddp st(1), st(0)                // m00*cos+m01*sin  m00*-sin  -sin  cos  sin
		fld DWORD PTR [eax+0+4]           // m01  m00*cos+m01*sin  m00*-sin  -sin  cos  sin
		fmul st(0), st(4)                 // m01*cos  m00*cos+m01*sin  m00*-sin  -sin  cos  sin
		fxch st(1)                        // m00*cos+m01*sin  m01*cos  m00*-sin  -sin  cos  sin
		fstp DWORD PTR [eax+0+0]          // m01*cos  m00*-sin  -sin  cos  sin
		faddp st(1), st(0)                // m01*cos+m00*-sin  -sin  cos  sin
		fld DWORD PTR [eax+16+0]          // m10  m01*cos+m00*-sin  -sin  cos  sin
		fmul st(0), st(3)                 // m10*cos  m01*cos+m00*-sin  -sin  cos  sin
		fxch st(1)                        // m01*cos+m00*-sin  m10*cos  -sin  cos  sin
		fstp DWORD PTR[eax+0+4]           // m10*cos  -sin  cos  sin
		fld DWORD PTR [eax+16+4]          // m11  m10*cos  -sin  cos  sin
		fmul st(0), st(4)                 // m11*sin  m10*cos  -sin  cos  sin
		fld DWORD PTR [eax+16+0]          // m10  m11*sin  m10*cos  -sin  cos  sin
		fmul st(0), st(3)                 // m10*-sin  m11*sin  m10*cos  -sin  cos  sin
		fxch st(2)                        // m10*cos  m11*sin  m10*-sin  -sin  cos  sin
		faddp st(1), st(0)                // m10*cos+m11*sin  m10*-sin  -sin  cos  sin
		fld DWORD PTR [eax+16+4]          // m11  m10*cos+m11*sin  m10*-sin  -sin  cos  sin
		fmul st(0), st(4)                 // m11*cos  m10*cos+m11*sin  m10*-sin  -sin  cos  sin
		fxch st(1)                        // m10*cos+m11*sin  m11*cos  m10*-sin  -sin  cos  sin
		fstp DWORD PTR [eax+16+0]         // m11*cos  m10*-sin  -sin  cos  sin
		faddp st(1), st(0)                // m11*cos+m10*-sin  -sin  cos  sin
		fld DWORD PTR[eax+16+16+0]        // m20  m11*cos+m10*-sin  -sin  cos  sin
		fmul st(0), st(3)                 // m20*cos  m11*cos+m10*-sin  -sin  cos  sin
		fxch st(1)                        // m11*cos+m10*-sin  m20*cos  -sin  cos  sin
		fstp DWORD PTR [eax+16+4]         // m20*cos  -sin  cos  sin
		fld DWORD PTR[eax+16+16+4]        // m21  m20*cos  -sin  cos  sin
		fmul st(0), st(4)                 // m21*sin  m20*cos  -sin  cos  sin
		fld DWORD PTR [eax+16+16+0]       // m20  m21*sin  m20*cos  -sin  cos  sin
		fmul st(0), st(3)                 // m20*-sin  m21*sin  m20*cos  -sin  cos  sin
		fxch st(2)                        // m20*cos  m21*sin  m20*-sin  -sin  cos  sin
		faddp st(1), st(0)                // m20*cos+m21*sin  m20*-sin  -sin  cos  sin
		fld DWORD PTR [eax+16+16+4]       // m21  m20*cos+m21*sin  m20*-sin  -sin  cos  sin
		fmul st(0), st(4)                 // m21*cos  m20*cos+m21*sin  m20*-sin  -sin  cos  sin
		fxch st(1)                        // m20*cos+m21*sin  m21*cos  m20*-sin  -sin  cos  sin
		fstp DWORD PTR [eax+16+16+0]      // m21*cos  m20*-sin  -sin  cos  sin
		faddp st(1), st(0)                // m21*cos+m20*-sin  -sin  cos  sin
		fxch st(3)                        // sin  -sin  cos  m21*cos+m20*-sin
		fstp st(0)                        // -sin  cos  m21*cos+m20*-sin
		fstp st(0)                        // cos  m21*cos+m20*-sin
		fstp st(0)                        // m21*cos+m20*-sin
		fstp DWORD PTR [eax+16+16+4]      // empty
	}

#else

	real s, c, a, b;

	s = sin(radians);
	c = cos(radians);

	a = matrix[0][0];
	b = matrix[0][1];
	matrix[0][0] = a *  c + b * s;
	matrix[0][1] = a * -s + b * c;

	a = matrix[1][0];
	b = matrix[1][1];
	matrix[1][0] = a *  c + b * s;
	matrix[1][1] = a * -s + b * c;

	a = matrix[2][0];
	b = matrix[2][1];
	matrix[2][0] = a *  c + b * s;
	matrix[2][1] = a * -s + b * c;

#endif
}

// ----------------------------------------------------------------------
#if !(PLATFORM == PLATFORM_WIN95 && USE_INLINE_ASM)
static Xform multiplyScratch;
#endif

// Multiply two transforms together and store the result in this transform
//
// Remarks:
//
//   The matrix to receive the result may be either one of the arguments to this function, in which case
//   a scratch Xform may used to protect ourselves from destroying a matrix in mid-computation.
//
//   This routine is highly optimized for the Win95 platform, and does not require a scratch Xform.

void Xform::multiply(
	const Xform &src1,  // the first matrix to multiply with
	const Xform &src2   // the second matrix to multiply by
	)
{
#if PLATFORM == PLATFORM_WIN95 && USE_INLINE_ASM

	// lint can't see the variable referenced in the inline assembly
	UNREF(src1);
	UNREF(src2);

	// assmebly by Intel, Chris Kastensmidt, Chris_Kastensmidt@ccm.fm.intel.com
 __asm
 {
		mov		eax, DWORD PTR src1
		mov		ebx, DWORD PTR src2
		mov		ecx, DWORD PTR this

		fld		DWORD PTR [ebx+12]	// d3
		fld		DWORD PTR [eax]
		fld		DWORD PTR [eax+16]
		fld		DWORD PTR [eax+32]	// c0 b0 a0 d3

		fmul	st(0), st(3)	// c0*d3
		fxch	st(2)		// a0 b0 c0*d3 d3
		fmul	st(0), st(3)	// a0*d3
		fxch				//  b0 a0*d3 c0*d3 d3
		fmulp	st(3), st(0) // b0*d3, pop

							//  All computations for a3/d3 finished

		fld		DWORD PTR [ebx+28]	// load e3
		fld		DWORD PTR [eax+4]		// load a1
		fld		DWORD PTR [ebx+44]	// load f3
		fld		DWORD PTR [eax+8]		// load a2
							//  a2 f3 a1 e3 a0*d3 c0*d3 b0*d3

		fmul	st(0), st(1)	// a2*f3
		fxch	st(2)		//  a1 f3 a2*f3 e3 a0*d3 c0*d3 b0*d3 
		fmul	st(0), st(3)	// a1*e3 f3 a2*f3 e3 a0*d3 c0*d3 b0*d3
		fadd	DWORD PTR [eax+12]	// a3+a1*e3 f3 a2*f3 e3 a0*d3 c0*d3 b0*d3
		faddp	st(2), st(0) // f3 a1*e3+a2*f3+a3 e3 a0*d3 c0*d3 b0*d3
		fxch	st(3)		//  a0*d3 a1*e3+a2*f3+a3 e3 f3 c0*d3 b0*d3
		faddp	st(1), st(0)	//  a0*d3+a1*e3+a2*f3+a3=matrix[0][3]

		fstp	DWORD PTR [ecx+12]	// e3 f3 c0*d3 b0*d3

		fld		DWORD PTR [eax+24]
		fld		DWORD PTR [eax+20]		// b1 b2 e3 f3 c0*d3 b0*d3
		fmul	st(0), st(2)	// b1*e3 b2 e3 f3 c0*d3 b0*d3
		fxch					// b2 b1*e3 e3 f3 c0*d3 b0*d3
		fmul	st(0), st(3)	// b2*f3 b1*e3 e3 f3 c0*d3 b0*d3
		fadd	DWORD PTR [eax+28]		// b2*f3+b3 b1*e3 e3 f3 c0*d3 b0*d3
		faddp	st(1), st(0)	// b1*e3+b2*f3+b3 e3 f3 c0*d3 b0*d3
		faddp	st(4), st(0)	// e3 f3 c0*d3 b0*d3+b1*e3+b2*f3+b3=matrix[1][3]
		fxch	st(3)
		fstp	DWORD PTR [ecx+28]	// f3 c0*d3 e3

		fld		DWORD PTR [eax+40]	
		fld		DWORD PTR [eax+36]	// c1 c2 f3 c0*d3 e3
		fmul	st(4), st(0)	// c1 c2 f3 c0*d3 c1*e3 
		fxch	st(1)		// c2 c1 f3 c0*d3 c1*e3
		fmul	st(2), st(0)	// c2 c1 c2*f3 c0*d3 c1*e3
		fld		DWORD PTR [eax+44]	// c3 c2 c1 c2*f3 c0*d3 c1*e3
		faddp	st(3), st(0)	// c2 c1 c2*f3+c3 c0*d3 c1*e3
		fxch	st(4)		//	c1*e3 c1 c2*f3+c3 c0*d3 c2
		faddp	st(2), st(0)	// c1 c1*e3+c2*f3+c3 c0*d3 c2
		fxch	st(1)		// c1*e3+c2*f3+c3 c1 c0*d3 c2
		faddp	st(2), st(0)	// c1 c0*d3+c1*e3+c2*f3+c3=matrix[2][3] c2
		fxch	
		fstp	DWORD PTR [ecx+44]	// c1 c2

		fld		DWORD PTR [ebx+40]	// f2 c1 c2
		fld		DWORD PTR [eax+32]	// c0 f2 c1 c2
		fld		DWORD PTR [ebx+24]
		fld		DWORD PTR [ebx+8]	// d2 e2 c0 f2 c1 c2
		fmul	st(0), st(2)	// d2*c0 e2 c0 f2 c1 c2
		fxch			// e2 d2*c0 c0 f2 c1 c2
		fmul	st(0), st(4)	// e2*c1 d2*c0 c0 f2 c1 c2
		fxch	st(3)	// f2 d2*c0 c0 e2*c1 c1 c2
		fmul	st(0), st(5)	// f2*c2 d2*c0 c0 e2*c1 c1 c2
		fld		DWORD PTR [ebx+40]	// f2 f2*c2 d2*c0 c0 e2*c1 c1 c2
		fxch	st(4)		// e2*c1 f2*c2 d2*c0 c0 f2 c1 c2
		faddp	st(1), st(0)	// c0*d2+c1*22 f2*c2 c0 f2 c1 c2
		faddp	st(1), st(0)	// c0*d2+c1*e2+c2*f2 c0 f2 c1 c2
		fstp	DWORD PTR [ecx+40]	//  store matrix[2][2]
							//  c0 f2 c1 c2
							//  Do not lose f2 and c2!!!

		fld		DWORD PTR [ebx+36]
		fld		DWORD PTR [ebx+36]	//  f1 f1 c0 f2 c1 c2
		fld		DWORD PTR [ebx+20]
		fld		DWORD PTR [ebx+4]		//  d1 e1 f1 f1 c0 f2 c1 c2

		fmul	st(0), st(4)	// d1*c0 e1 f1 f1 c0 f2 c1 c2
		fxch	st(1)		// e1 c0*d1 f1 f1 c0 f2 c1 c2
		fmul	st(0), st(6)	// c1*e1 c0*d1 f1 f1 c0 f2 c1 c2
		fxch	st(2)		// f1 c1*e1 c0*d1 f1 c0 f2 c1 c2
		fmul	st(0), st(7)	// f1*c2 c1*e1 c0*d1 f1 c0 f2 c1 c2
		faddp	st(2), st(0)	// c1*e1 c0*d1+c2*f1 f1 c0 f2 c1 c2
		faddp	st(1), st(0)	// c0*d1+c1*e1+c2*f1 f1 c0 f2 c1 c2
		fstp	DWORD PTR [ecx+36]	// store matrix[2][1]
							// f1 c0 f2 c1 c2
							// Do not lose f1 and c1!!!

		fld		DWORD PTR [ebx+16]
		fld		DWORD PTR [ebx+0]		// d0 e0 f1 c0 f2 c1 c2
		fmul	st(3), st(0)	// d0 e0 f1 c0*d0 f2 c1 c2 (finished with c0)
		fxch	st(1)		// e0 d0 f1 c0*d0 f2 c1 c2
		fmul	st(5), st(0)	// e0 d0 f1 c0*d0 f2 c1*e0 c2 (finished with c1)
		fld		DWORD PTR [ebx+32]	// f0 e0 d0 f1 c0*d0 f2 c1*e0 c2
		fmul	st(7), st(0)  // f0 e0 d0 f1 c0*d0 f2 c1*e0 c2*f0 (finished with c2)
		fxch	st(7)		// c2*f0 e0 d0 f1 c0*d0 f2 c1*e0 f0
		faddp	st(4), st(0) // e0 d0 f1 c0*d0+c2*f0 f2 c1*e0 f0
		fxch	st(5)		// c1*e0 d0 f1 c0*d0+c2*f0 f2 e0 f0
		faddp	st(3), st(0)  // d0 f1 c0*d0+c2*f0 f2 e0 f0
		fxch	st(2)		// c0*d0+c2*f0 f1 d0 f2 e0 f0
		fstp	DWORD PTR [ecx+32]	// store matrix[2][0]
							// f1 d0 f2 e0 f0
							// Currently, do not lose f0, f1, f2 only

		fld		DWORD PTR [eax+24]
		fld		DWORD PTR [eax+20]
		fld		DWORD PTR [eax+16]	//	b0 b1 b2 f1 d0 f2 e0 f0

		fmul	st(4), st(0)	// b0 b1 b2 f1 b0*d0 f2 e0 f0
		fxch	st(6)		// e0 b1 b2 f1 b0*d0 f2 b0 f0
		fmulp	st(1),st(0)		// e0*b1 b2 f1 b0*d0 f2 b0 f0
		fxch	st(1)		// b2 e0*b1 f1 b0*d0 f2 b0 f0
		fmul	st(0),st(6)	// b2*f0 b1*e0 f1 b0*d0 f2 b0 f0
		faddp	st(3),st(0)	// b1*e0 f1 b0*d0+b2*f0 f2 b0 f0
		faddp	st(2),st(0)	// f1 b0*d0+b1*e0+b2*f0 f2 b0 f0
		fld		DWORD PTR [ebx+16]	// e0 f1 b0*d0+b1*e0+b2*f0 f2 b0 f0
		fxch	st(2)		// b0*d0+b1*e0+b2*f0 f1 e0 f2 b0 f0
		fstp	DWORD PTR [ecx+16]	// store matrix[1][0]
							// f1 e0 f2 b0 f0
							// Currently, do not lose f0, f1, f2, e0, b0

		fld		DWORD PTR [eax+0]
		fld		DWORD PTR [eax+4]
		fld		DWORD PTR [eax+8]		//	a2 a1 a0 f1 e0 f2 b0 f0
		fmul	st(7),st(0)	// a2 a1 a0 f1 e0 f2 b0 a2*f0  (f0 finished)
		fxch	st(4)		// e0 a1 a0 f1 a2 f2 b0 a2*f0
		fmul	st(0),st(1)	// e0*a1 a1 a0 f1 a2 f2 b0 a2*f0  (e0 finished)
		faddp	st(7),st(0)	// a1 a0 f1 a2 f2 b0 a1*e0+a2*f0
		fld		DWORD PTR [ebx+0]		// d0 a1 a0 f1 a2 f2 b0 a1*e0+a2*f0
		fmul	st(0),st(2)	// d0*a0 a1 a0 f1 a2 f2 b0 a1*e0+a2*f0 (d0 finished)
		faddp	st(7),st(0)	// a1 a0 f1 a2 f2 b0 a0*d0+a1*e0+a2*f0
		fxch	st(6)		
		fstp	DWORD PTR [ecx]		// store matrix[0][0]
							// a0 f1 a2 f2 b0 a1	
							// Currently, do not lose f1, f2, a0, b0

		fld		DWORD PTR [ebx+4]		
		fld		DWORD PTR [ebx+20]	// e1 d1 a0 f1 a2 f2 b0 a1
		fmul	st(0),st(7)	// a1*e1 d1 a0 f1 a2 f2 b0 a1
		fxch	st(1)		// d1 a1*e1 a0 f1 a2 f2 b0 a1
		fmul	st(0),st(2)	// a0*d1 a1*e1 a0 f1 a2 f2 b0 a1
		fxch	st(4)		// a2 a1*e1 a0 f1 a0*d1 f2 b0 a1
		fmul	st(0),st(3)	// a2*f1 a1*e1 a0 f1 a0*d1 f2 b0 a1
		faddp	st(4),st(0)	// a1*e1 a0 f1 a0*d1+a2*f1 f2 b0 a1
		faddp	st(3),st(0) // a0 f1 a0*d1+a1*e1+a2*f1 f2 b0 a1
		fld		DWORD PTR [ebx+4]		// d1 a0 f1 a0*d1+a1*e1+a2*f1 f2 b0 a1
		fxch	st(3)		// a0*d1+a1*e1+a2*f1 a0 f1 d1 f2 b0 a1
		fstp	DWORD PTR [ecx+4]		// store matrix[0][1]
							// a0 f1 d1 f2 b0 a1
							// Currently, do not lose f1, f2, a0, b0, a1, d1

		fxch	st(2)		// d1 f1 a0 f2 b0 a1
		fmul	st(0),st(4)	// b0*d1 f1 a0 f2 b0 a1 (d1 finished)
		fld		DWORD PTR [eax+20]	// b1 b0*d1 f1 a0 f2 b0 a1
		fld		DWORD PTR [ebx+20]	// e1 b1 b0*d1 f1 a0 f2 b0 a1
		fmul	st(0),st(1)	// e1*b1 b1 b0*d1 f1 a0 f2 b0 a1	(e1 finished)
		faddp	st(2),st(0)	// b1 b0*d1+b1*e1 f1 a0 f2 b0 a1
		fld		DWORD PTR [eax+24]	// b2 b1 b0*d1+b1*e1 f1 a0 f2 b0 a1
		fmul	st(3),st(0)	// b2 b1 b0*d1+b1*e1 b2*f1 a0 f2 b0 a1 (f1 finished)
		fxch	st(3)		// b2*f1 b1 b0*d1+b1*e1 b2 a0 f2 b0 a1
		faddp	st(2),st(0)	// b1 b0*d1+b1*e1+b2*f1 b2 a0 f2 b0 a1
		fxch	st(1)		// b0*d1+b1*e1+b2*f1 b1 b2 a0 f2 b0 a1
		fstp	DWORD PTR [ecx+20]	// store matrix[1][1]
							// b1 b2 a0 f2 b0 a1
							// currently, do not lose f2, a0, b0, a1, b1

		fld		DWORD PTR [ebx+24]	
		fld		DWORD PTR [ebx+8]		// d2 e2 b1 b2 a0 f2 b0 a1
		fmul	st(6),st(0)	// d2 e2 b1 b2 a0 f2 b0*d2 a1 (b0 finished)
		fxch	st(1)		// e2 d2 b1 b2 a0 f2 b0*d2 a1
		fmul	st(2),st(0)	// e2 d2 b1*e2 b2 a0 f2 b0*d2 a1 (b1 finished)
		fxch	st(3)		// b2 d2 b1*e2 e2 a0 f2 b0*d2 a1
		fmul	st(0),st(5)	// b2*f2 d2 b1*e2 e2 a0 f2 b0*d2 a1 (b2 finished)
		faddp	st(6),st(0)	// d2 b1*e2 e2 a0 f2 b0*d2+b2*f2 a1
		fxch	st(5)		// b0*d2+b2*f2 b1*e2 e2 a0 f2 d2 a1
		faddp	st(1),st(0)
		fstp	DWORD PTR [ecx+24]	// store matrix[1][2]
							// e2 a0 f2 d2 a1	
							// Only six variables left!

		fld		DWORD PTR [eax+8]		// a2 e2 a0 f2 d2 a1
		fmulp	st(3),st(0)	// e2 a0 a2*f2 d2 a1
		fmulp	st(4),st(0)	// a0 a2*f2 d2 a1*e2
		fmulp	st(2),st(0) // a2*f2 a0*d2 a1*e2
		faddp	st(2),st(0)
		faddp	st(1),st(0) // a0*d2+a1*e2+a2*f2
		fstp	DWORD PTR [ecx+8]		// store matrix[0][2] - FP stack is cleared
	}

#else

	const Xform *s1, *s2;

	if (this == &src1)
	{
		multiplyScratch = *this;
		s1 = &multiplyScratch;
		s2 = &src2;
	}
	else
		if (this == &src2)
		{
			multiplyScratch = *this;
			s1 = &src1;
			s2 = &multiplyScratch;
		}
		else
		{
			s1 = &src1;
			s2 = &src2;
		}

	for (int row = 0; row < 3; row++)
	{
		for (int col = 0; col < 3; col++)
			matrix[row][col] =
				s1->matrix[row][0] * s2->matrix[0][col] +
				s1->matrix[row][1] * s2->matrix[1][col] +
				s1->matrix[row][2] * s2->matrix[2][col];

		matrix[row][col] =
			s1->matrix[row][0] * s2->matrix[0][col] +
			s1->matrix[row][1] * s2->matrix[1][col] +
			s1->matrix[row][2] * s2->matrix[2][col] +
			s1->matrix[row][3];
	}

#endif
}

// ----------------------------------------------------------------------

void Xform::makeScale(const Vector &scale)
{
	matrix[0][0] = scale.x;
	matrix[1][1] = scale.y;
	matrix[2][2] = scale.z;
}

// ----------------------------------------------------------------------
// Invert a transform
//
// Remarks:
//
//   The inverse of the upper 3x3 of a purely rotational matrix is simply its transpose.

void Xform::invert(
	const Xform &m  // the transform to invert
	)
{
	int  row, col;
	real x, y, z;

	DEBUG_FATAL(this == &m, ("Xform::invert from self"));

	// calculate the new positional offset
	x = -m.matrix[0][3];
	y = -m.matrix[1][3];
	z = -m.matrix[2][3];
	matrix[0][3] = m.matrix[0][0]*x + m.matrix[1][0]*y + m.matrix[2][0]*z;
	matrix[1][3] = m.matrix[0][1]*x + m.matrix[1][1]*y + m.matrix[2][1]*z;
	matrix[2][3] = m.matrix[0][2]*x + m.matrix[1][2]*y + m.matrix[2][2]*z;

	// transpose the upper 3x3 matrix
	for (row = 0; row < 3; row++)
		for (col = 0; col < 3; col++)
			matrix[row][col] = m.matrix[col][row];
}

// ------------------------------------------------------------------------
// Translate a vector from transform space to parent space
//
// Remarks:
//
//   This routine simply adds this transform's x, y, and z position to the source vector.

void Xform::translate(
	const Vector &src, // source vector in transform space
	Vector       *dest // vector to receive the parent space result
	) const
{
#if PLATFORM == PLATFORM_WIN95

	// lint can't see the variable referenced in the inline assembly
	UNREF(src);
	UNREF(dest);

	__asm
	{
		mov esi, src
		mov eax, this
		mov edi, dest

		fld DWORD PTR [esi]               // sx
		fadd DWORD PTR [eax+12]           // sx+m03
		fld DWORD PTR [esi+4]             // sy      sx+m03
		fadd DWORD PTR [eax+16+12]        // sy+m13  sx+m03
		fld DWORD PTR [esi+8]             // sz      sy+m13  sx+m03
		fadd DWORD PTR [eax+16+16+12]     // sz+m23  sy+m13  sx+m03
		fxch st(2)                        // sx+m03  sy+m13  sz+m13
		fstp DWORD PTR [edi]              // sy+m13  sz+m13
		fstp DWORD PTR [edi+4]            // sz+m13
		fstp DWORD PTR [edi+8]            // empty
	}

#else

	dest->x = src.x + matrix[0][3]; 
	dest->y = src.y + matrix[1][3]; 
	dest->z = src.z + matrix[2][3]; 

#endif
}

// ----------------------------------------------------------------------
// Rotate the source vector from transform's space into parent space
//
// Remarks:
//
//   Simple rotation is appropriate for tranformation of directional vectors, for example normal vectors.
//
//   This routine is highly optimized for the Win95 platform.
//
// See also:
//
//   rotateTranslate(), inverseRotate()

void Xform::rotate(
	const Vector &src,  // source directional vector in transform space
	Vector       *dest  // vector to receive the parent space result
	) const
{
#if PLATFORM == PLATFORM_WIN95

	// lint can't see the variable referenced in the inline assembly
	UNREF(src);
	UNREF(dest);

	__asm
	{
		mov esi, src
		mov eax, this
		mov edi, dest
		                             // s0                   s1                   s2                   s3    s4    s5        

		fld DWORD PTR [esi+0]        // sX
		fld DWORD PTR [esi+4]        // sY                   sX
		fld DWORD PTR [esi+8]        // sZ                   sY                   sX
		fld DWORD PTR [eax+0]        // m00                  sZ                   sY                   sX
		fmul st(0), st(3)            // m00*sX               sZ                   sY                   sX
		fld DWORD PTR [eax+4]        // m01                  m00*sX               sZ                   sY    sX
		fmul st(0), st(3)            // m01*sY               m00*sX               sZ                   sY    sX
		fld DWORD PTR [eax+8]        // m02                  m01*sY               m00*sX               sZ    sY    sX
		fmul st(0), st(3)            // m02*sZ               m01*sY               m00*sX               sZ    sY    sX 
		fxch st(2)                   // m00*sX               m01*sY               m02*sZ               sZ    sY    sX
		faddp st(1), st(0)           // m00*sX+m01*sY        m02*sZ               sZ                   sY    sX
		fld DWORD PTR [eax+16+0]     // m10                  m00*sX+m01*sY        m02*sZ               sZ    sY    sX
		fmul st(0), st(5)            // m10*sX               m00*sX+m01*sY        m02*sZ               sZ    sY    sX
		fxch st(2)                   // m02*sZ               m00*sX+m01*sY        m10*sX               sZ    sY    sX
		faddp st(1), st(0)           // m00*sX+m01*sY+m02*sZ m10*sX               sZ                   sY    sX
		fld DWORD PTR [eax+16+4]     // m11                  m00*sX+m01*sY+m02*sZ m10*sX               sZ    sY    sX
		fmul st(0), st(4)            // m11*sy               m00*sX+m01*sY+m02*sZ m10*sX               sZ    sY    sX
		fxch st(1)                   // m00*sX+m01*sY+m02*sZ m11*sy               m10*sX               sZ    sY    sX
		fstp DWORD PTR [edi+0]       // m11*sy               m10*sX               sZ                   sY    sX
		fld DWORD PTR [eax+16+8]     // m12                  m11*sy               m10*sX               sZ    sY    sX
		fmul st(0), st(3)            // m12*sZ               m11*sy               m10*sX               sZ    sY    sX
		fxch st(2)                   // m10*sX               m11*sy               m12*sZ               sZ    sY    sX
		faddp st(1), st(0)           // m10*sX+m11*sy        m12*sZ               sZ                   sY    sX
		fld DWORD PTR [eax+16+16+0]  // m20                  m10*sX+m11*sy        m12*sZ               sZ    sY    sX
		fmul st(0), st(5)            // m20*sx               m10*sX+m11*sy        m12*sZ               sZ    sY    sX
		fxch st(2)                   // m12*sZ               m10*sX+m11*sy        m20*sx               sZ    sY    sX
		faddp st(1), st(0)           // m10*sX+m11*sy+m12*sZ m20*sx               sZ                   sY    sX
		fld DWORD PTR [eax+16+16+4]  // m21                  m10*sX+m11*sy+m12*sZ m20*sx               sZ    sY    sX
		fmul st(0), st(4)            // m21*sy               m10*sX+m11*sy+m12*sZ m20*sx               sZ    sY    sX
		fxch st(1)                   // m10*sX+m11*sy+m12*sZ m21*sy               m20*sx               sZ    sY    sX
		fstp DWORD PTR [edi+4]       // m21*sy               m20*sx               sZ                   sY    sX
		fld DWORD PTR[eax+16+16+8]   // m22                  m21*sy               m20*sx               sZ    sY    sX
		fmul st(0), st(3)            // m22*sz               m21*sy               m20*sx               sZ    sY    sX
		fxch st(2)                   // m20*sx               m21*sy               m22*sz               sZ    sY    sX
		faddp st(1), st(0)           // m20*sx+m21*sy        m22*sz               sZ                   sY    sX
		fxch st(2)                   // sz                   m20*sx+m21*sy        m22*sz               sY    sX
		fstp st(0)                   // m20*sx+m21*sy        m22*sz               sY                   sX
		faddp st(1), st(0)           // m20*sx+m21*sy+m22*sz sY                   sX
		fxch st(2)                   // sx                   sy                   m20*sx+m21*sy+m22*sz
		fstp st(0)                   // sy                   m20*sx+m21*sy+m22*sz
		fstp st(0)                   // m20*sx+m21*sy+m22*sz
		fstp DWORD PTR [edi+8]       // empty
	}

#else

	real x, y, z;

	x = src.x;
	y = src.y;
	z = src.z;

	dest->x = matrix[0][0]*x + matrix[0][1]*y + matrix[0][2]*z;
	dest->y = matrix[1][0]*x + matrix[1][1]*y + matrix[1][2]*z;
	dest->z = matrix[2][0]*x + matrix[2][1]*y + matrix[2][2]*z;

#endif
}

// ------------------------------------------------------------------------
// Rotate and translate the source vector from transform space into parent space
//
// Remarks:
//
//   Rotation and translation is appropriate for the transformation of positional vectors, for example particular
//   points in space.
//
//   This routine is highly optimized for the Win95 platform.
//
// See also:
//
//   rotate(), inverseRotateTranslate()

void Xform::rotateTranslate(
	const Vector &src,  // source positional vector in transform space
	Vector       *dest  // vector to receive the parent space result
	) const
{
#if PLATFORM == PLATFORM_WIN95

	// lint can't see the variable referenced in the inline assembly
	UNREF(src);
	UNREF(dest);

	__asm
	{
		mov esi, src
		mov eax, this
		mov edi, dest
		                             // s0                   s1                   s2                   s3    s4    s5        

		fld DWORD PTR [esi+0]        // sX
		fld DWORD PTR [esi+4]        // sY                   sX
		fld DWORD PTR [esi+8]        // sZ                   sY                   sX
		fld DWORD PTR [eax+0]        // m00                  sZ                   sY                   sX
		fmul st(0), st(3)            // m00*sX               sZ                   sY                   sX
		fld DWORD PTR [eax+4]        // m01                  m00*sX               sZ                   sY    sX
		fmul st(0), st(3)            // m01*sY               m00*sX               sZ                   sY    sX
		fld DWORD PTR [eax+8]        // m02                  m01*sY               m00*sX               sZ    sY    sX
		fmul st(0), st(3)            // m02*sZ               m01*sY               m00*sX               sZ    sY    sX 
		fxch st(2)                   // m00*sX               m01*sY               m02*sZ               sZ    sY    sX
		faddp st(1), st(0)           // m00*sX+m01*sY        m02*sZ               sZ                   sY    sX
		fld DWORD PTR [eax+16+0]     // m10                  m00*sX+m01*sY        m02*sZ               sZ    sY    sX
		fmul st(0), st(5)            // m10*sX               m00*sX+m01*sY        m02*sZ               sZ    sY    sX
		fxch st(2)                   // m02*sZ               m00*sX+m01*sY        m10*sX               sZ    sY    sX
		faddp st(1), st(0)           // m00*sX+m01*sY+m02*sZ m10*sX               sZ                   sY    sX
		fld DWORD PTR [eax+16+4]     // m11                  m00*sX+m01*sY+m02*sZ m10*sX               sZ    sY    sX
		fmul st(0), st(4)            // m11*sy               m00*sX+m01*sY+m02*sZ m10*sX               sZ    sY    sX
		fxch st(1)                   // m00*sX+m01*sY+m02*sZ m11*sy               m10*sX               sZ    sY    sX
		fadd DWORD PTR [eax+12]      // dx                   m11*sy               m10*sX               sZ    sY    sX
		fstp DWORD PTR [edi+0]       // m11*sy               m10*sX               sZ                   sY    sX
		fld DWORD PTR [eax+16+8]     // m12                  m11*sy               m10*sX               sZ    sY    sX
		fmul st(0), st(3)            // m12*sZ               m11*sy               m10*sX               sZ    sY    sX
		fxch st(2)                   // m10*sX               m11*sy               m12*sZ               sZ    sY    sX
		faddp st(1), st(0)           // m10*sX+m11*sy        m12*sZ               sZ                   sY    sX
		fld DWORD PTR [eax+16+16+0]  // m20                  m10*sX+m11*sy        m12*sZ               sZ    sY    sX
		fmul st(0), st(5)            // m20*sx               m10*sX+m11*sy        m12*sZ               sZ    sY    sX
		fxch st(2)                   // m12*sZ               m10*sX+m11*sy        m20*sx               sZ    sY    sX
		faddp st(1), st(0)           // m10*sX+m11*sy+m12*sZ m20*sx               sZ                   sY    sX
		fld DWORD PTR [eax+16+16+4]  // m21                  m10*sX+m11*sy+m12*sZ m20*sx               sZ    sY    sX
		fmul st(0), st(4)            // m21*sy               m10*sX+m11*sy+m12*sZ m20*sx               sZ    sY    sX
		fxch st(1)                   // m10*sX+m11*sy+m12*sZ m21*sy               m20*sx               sZ    sY    sX
		fadd DWORD PTR [eax+16+12]   // dy                   m21*sy               m20*sx               sZ    sY    sX
		fstp DWORD PTR [edi+4]       // m21*sy               m20*sx               sZ                   sY    sX
		fld DWORD PTR[eax+16+16+8]   // m22                  m21*sy               m20*sx               sZ    sY    sX
		fmul st(0), st(3)            // m22*sz               m21*sy               m20*sx               sZ    sY    sX
		fxch st(2)                   // m20*sx               m21*sy               m22*sz               sZ    sY    sX
		faddp st(1), st(0)           // m20*sx+m21*sy        m22*sz               sZ                   sY    sX
		fxch st(2)                   // sz                   m20*sx+m21*sy        m22*sz               sY    sX
		fstp st(0)                   // m20*sx+m21*sy        m22*sz               sY                   sX
		faddp st(1), st(0)           // m20*sx+m21*sy+m22*sz sY                   sX
		fxch st(2)                   // sx                   sy                   m20*sx+m21*sy+m22*sz
		fstp st(0)                   // sy                   m20*sx+m21*sy+m22*sz
		fstp st(0)                   // m20*sx+m21*sy+m22*sz
		fadd DWORD PTR [eax+16+16+12]// dz
		fstp DWORD PTR [edi+8]       // empty
	}

#else

	real x, y, z;

	x = src.x;
	y = src.y;
	z = src.z;

	dest->x = matrix[0][0]*x + matrix[0][1]*y + matrix[0][2]*z + matrix[0][3];
	dest->y = matrix[1][0]*x + matrix[1][1]*y + matrix[1][2]*z + matrix[1][3];
	dest->z = matrix[2][0]*x + matrix[2][1]*y + matrix[2][2]*z + matrix[2][3];

#endif
}

// ------------------------------------------------------------------------
// Rotate the source vector from parent space into transform space
//
// Remarks:
//
//   Simple rotation is appropriate for tranformation of directional vectors, for example normal vectors.
//
//   This routine is highly optimized for the Win95 platform.
//
// See also:
//
//   inverseRotateTranslate(), rotate()

void Xform::inverseRotate(
	const Vector &src,  // source directional vector in parent space
	Vector       *dest  // vector to receive transform space result
	) const
{
#if PLATFORM == PLATFORM_WIN95

	// lint can't see the variable referenced in the inline assembly
	UNREF(src);
	UNREF(dest);

	__asm
	{
		mov esi, src
		mov eax, this
		mov edi, dest
		                             // s0                   s1                   s2                   s3    s4    s5        
	
		fld DWORD PTR [esi+0]        // sX
		fld DWORD PTR [esi+4]        // sY                   sX
		fld DWORD PTR [esi+8]        // sZ                   sY                   sX
		fld DWORD PTR [eax+0+0]      // m00                  sZ                   sY                   sX
		fmul st(0), st(3)            // m00*sX               sZ                   sY                   sX
		fld DWORD PTR [eax+16+0]     // m10                  m00*sX               sZ                   sY    sX
		fmul st(0), st(3)            // m10*sY               m00*sX               sZ                   sY    sX
		fld DWORD PTR [eax+16+16+0]  // m20                  m10*sY               m00*sX               sZ    sY    sX
		fmul st(0), st(3)            // m20*sZ               m10*sY               m00*sX               sZ    sY    sX 
		fxch st(2)                   // m00*sX               m10*sY               m20*sZ               sZ    sY    sX
		faddp st(1), st(0)           // m00*sX+m10*sY        m20*sZ               sZ                   sY    sX
		fld DWORD PTR [eax+0+4]      // m01                  m00*sX+m10*sY        m20*sZ               sZ    sY    sX
		fmul st(0), st(5)            // m01*sX               m00*sX+m10*sY        m20*sZ               sZ    sY    sX
		fxch st(2)                   // m20*sZ               m00*sX+m10*sY        m01*sX               sZ    sY    sX
		faddp st(1), st(0)           // m00*sX+m10*sY+m20*sZ m01*sX               sZ                   sY    sX
		fld DWORD PTR [eax+16+4]     // m11                  m00*sX+m10*sY+m20*sZ m01*sX               sZ    sY    sX
		fmul st(0), st(4)            // m11*sy               m00*sX+m10*sY+m20*sZ m01*sX               sZ    sY    sX
		fxch st(1)                   // m00*sX+m10*sY+m20*sZ m11*sy               m01*sX               sZ    sY    sX
		fstp DWORD PTR [edi+0]       // m11*sy               m01*sX               sZ                   sY    sX
		fld DWORD PTR [eax+16+16+4]  // m21                  m11*sy               m01*sX               sZ    sY    sX
		fmul st(0), st(3)            // m21*sZ               m11*sy               m01*sX               sZ    sY    sX
		fxch st(2)                   // m01*sX               m11*sy               m21*sZ               sZ    sY    sX
		faddp st(1), st(0)           // m01*sX+m11*sy        m21*sZ               sZ                   sY    sX
		fld DWORD PTR [eax+0+8]      // m02                  m01*sX+m11*sy        m21*sZ               sZ    sY    sX
		fmul st(0), st(5)            // m02*sx               m01*sX+m11*sy        m21*sZ               sZ    sY    sX
		fxch st(2)                   // m21*sZ               m01*sX+m11*sy        m02*sx               sZ    sY    sX
		faddp st(1), st(0)           // m01*sX+m11*sy+m21*sZ m02*sx               sZ                   sY    sX
		fld DWORD PTR [eax+16+8]     // m12                  m01*sX+m11*sy+m21*sZ m02*sx               sZ    sY    sX
		fmul st(0), st(4)            // m12*sy               m01*sX+m11*sy+m21*sZ m02*sx               sZ    sY    sX
		fxch st(1)                   // m01*sX+m11*sy+m21*sZ m12*sy               m02*sx               sZ    sY    sX
		fstp DWORD PTR [edi+4]       // m12*sy               m02*sx               sZ                   sY    sX
		fld DWORD PTR[eax+16+16+8]   // m22                  m12*sy               m02*sx               sZ    sY    sX
		fmul st(0), st(3)            // m22*sz               m12*sy               m02*sx               sZ    sY    sX
		fxch st(2)                   // m02*sx               m12*sy               m22*sz               sZ    sY    sX
		faddp st(1), st(0)           // m02*sx+m12*sy        m22*sz               sZ                   sY    sX
		fxch st(2)                   // sz                   m02*sx+m12*sy        m22*sz               sY    sX
		fstp st(0)                   // m02*sx+m12*sy        m22*sz               sY                   sX
		faddp st(1), st(0)           // m02*sx+m12*sy+m22*sz sY                   sX
		fxch st(2)                   // sx                   sy                   m02*sx+m12*sy+m22*sz
		fstp st(0)                   // sy                   m02*sx+m12*sy+m22*sz
		fstp st(0)                   // m02*sx+m12*sy+m22*sz
		fstp DWORD PTR [edi+8]       // empty
	}

#else

	real x, y, z;

	x = src.x;
	y = src.y;
	z = src.z;

	dest->x = matrix[0][0]*x + matrix[1][0]*y + matrix[2][0]*z;
	dest->y = matrix[0][1]*x + matrix[1][1]*y + matrix[2][1]*z;
	dest->z = matrix[0][2]*x + matrix[1][2]*y + matrix[2][2]*z;

#endif
}

// ------------------------------------------------------------------------
// Rotate and translate the source vector from parent space into transform space
//
// Remarks:
//
//   Rotation and translation is appropriate for tranformation of positional vectors, for example specified points in
//   space.
//
//   This routine is highly optimized for the Win95 platform.
//
// See also:
//
//   inverseRotate(), rotateTranslate()

void Xform::inverseRotateTranslate(
	const Vector &src,   // source positional vector in parent space
	Vector       *dest   // vector to receive the transform space result
	) const
{
#if PLATFORM == PLATFORM_WIN95

	// lint can't see the variable referenced in the inline assembly
	UNREF(src);
	UNREF(dest);

	__asm
	{
		mov esi, src
		mov eax, this
		mov edi, dest
		                             // s0                   s1                   s2                   s3    s4    s5        

		fld DWORD PTR [esi+0]        // sX
		fsub DWORD PTR[eax+12]
		fld DWORD PTR [esi+4]        // sY                   sX
		fsub DWORD PTR[eax+16+12]
		fld DWORD PTR [esi+8]        // sZ                   sY                   sX
		fsub DWORD PTR[eax+16+16+12]

		fld DWORD PTR [eax+0+0]      // m00                  sZ                   sY                   sX
		fmul st(0), st(3)            // m00*sX               sZ                   sY                   sX
		fld DWORD PTR [eax+16+0]     // m10                  m00*sX               sZ                   sY    sX
		fmul st(0), st(3)            // m10*sY               m00*sX               sZ                   sY    sX
		fld DWORD PTR [eax+16+16+0]  // m20                  m10*sY               m00*sX               sZ    sY    sX
		fmul st(0), st(3)            // m20*sZ               m10*sY               m00*sX               sZ    sY    sX 
		fxch st(2)                   // m00*sX               m10*sY               m20*sZ               sZ    sY    sX
		faddp st(1), st(0)           // m00*sX+m10*sY        m20*sZ               sZ                   sY    sX
		fld DWORD PTR [eax+0+4]      // m01                  m00*sX+m10*sY        m20*sZ               sZ    sY    sX
		fmul st(0), st(5)            // m01*sX               m00*sX+m10*sY        m20*sZ               sZ    sY    sX
		fxch st(2)                   // m20*sZ               m00*sX+m10*sY        m01*sX               sZ    sY    sX
		faddp st(1), st(0)           // m00*sX+m10*sY+m20*sZ m01*sX               sZ                   sY    sX
		fld DWORD PTR [eax+16+4]     // m11                  m00*sX+m10*sY+m20*sZ m01*sX               sZ    sY    sX
		fmul st(0), st(4)            // m11*sy               m00*sX+m10*sY+m20*sZ m01*sX               sZ    sY    sX
		fxch st(1)                   // m00*sX+m10*sY+m20*sZ m11*sy               m01*sX               sZ    sY    sX
		fstp DWORD PTR [edi+0]       // m11*sy               m01*sX               sZ                   sY    sX
		fld DWORD PTR [eax+16+16+4]  // m21                  m11*sy               m01*sX               sZ    sY    sX
		fmul st(0), st(3)            // m21*sZ               m11*sy               m01*sX               sZ    sY    sX
		fxch st(2)                   // m01*sX               m11*sy               m21*sZ               sZ    sY    sX
		faddp st(1), st(0)           // m01*sX+m11*sy        m21*sZ               sZ                   sY    sX
		fld DWORD PTR [eax+0+8]      // m02                  m01*sX+m11*sy        m21*sZ               sZ    sY    sX
		fmul st(0), st(5)            // m02*sx               m01*sX+m11*sy        m21*sZ               sZ    sY    sX
		fxch st(2)                   // m21*sZ               m01*sX+m11*sy        m02*sx               sZ    sY    sX
		faddp st(1), st(0)           // m01*sX+m11*sy+m21*sZ m02*sx               sZ                   sY    sX
		fld DWORD PTR [eax+16+8]     // m12                  m01*sX+m11*sy+m21*sZ m02*sx               sZ    sY    sX
		fmul st(0), st(4)            // m12*sy               m01*sX+m11*sy+m21*sZ m02*sx               sZ    sY    sX
		fxch st(1)                   // m01*sX+m11*sy+m21*sZ m12*sy               m02*sx               sZ    sY    sX
		fstp DWORD PTR [edi+4]       // m12*sy               m02*sx               sZ                   sY    sX
		fld DWORD PTR[eax+16+16+8]   // m22                  m12*sy               m02*sx               sZ    sY    sX
		fmul st(0), st(3)            // m22*sz               m12*sy               m02*sx               sZ    sY    sX
		fxch st(2)                   // m02*sx               m12*sy               m22*sz               sZ    sY    sX
		faddp st(1), st(0)           // m02*sx+m12*sy        m22*sz               sZ                   sY    sX
		fxch st(2)                   // sz                   m02*sx+m12*sy        m22*sz               sY    sX
		fstp st(0)                   // m02*sx+m12*sy        m22*sz               sY                   sX
		faddp st(1), st(0)           // m02*sx+m12*sy+m22*sz sY                   sX
		fxch st(2)                   // sx                   sy                   m02*sx+m12*sy+m22*sz
		fstp st(0)                   // sy                   m02*sx+m12*sy+m22*sz
		fstp st(0)                   // m02*sx+m12*sy+m22*sz
		fstp DWORD PTR [edi+8]       // empty
	}

#else

	real x, y, z;

	x = src.x - matrix[0][3];
	y = src.y - matrix[1][3];
	z = src.z - matrix[2][3];

	dest->x = matrix[0][0]*x + matrix[1][0]*y + matrix[2][0]*z;
	dest->y = matrix[0][1]*x + matrix[1][1]*y + matrix[2][1]*z;
	dest->z = matrix[0][2]*x + matrix[1][2]*y + matrix[2][2]*z;

#endif
}

// ------------------------------------------------------------------------
// Rotate and translate an array of source vectors from transform space into parent space
//
// Remarks:
//
//   Rotation and translation is appropriate for the transformation of positional vectors, for example particular
//   points in space.
//
//   Unlike many other routines in this class, this routine is NOT highly optimized for the Win95 platform.
//
// See also:
//
//   rotate(), inverseRotateTranslate()

void Xform::rotateTranslate(
	const Vector *src,   // source array of positional vectors in object space
	Vector       *dest,  // array of vectors to receive the parent space results
	int cnt              // number of vectors to transform
	) const
{
	real x, y, z;
	
	for (; cnt--; src++, dest++)
	{
		x = src->x;
		y = src->y;
		z = src->z;
		
		dest->x = matrix[0][0]*x + matrix[0][1]*y + matrix[0][2]*z + matrix[0][3];
		dest->y = matrix[1][0]*x + matrix[1][1]*y + matrix[1][2]*z + matrix[1][3];
		dest->z = matrix[2][0]*x + matrix[2][1]*y + matrix[2][2]*z + matrix[2][3];
	}
}

// ------------------------------------------------------------------------
// Rotate and translate a specified plane from transform space to parent space
//
// Remarks:
//
//   This routine will rotate the plane normal from transform space to parent space.  It will then compute a point
//   on the original plane, rotate and translate that into parent space, and use that point to recalculate the 
//   D plane coefficient for the plane in parent space.

void Xform::rotateTranslatePlane(
	const Vector &srcNormal,   // normal of plane in transform space
	real          srcPlaneD,   // plane D coefficient in transform space
	Vector       *destNormal,  // vector to receive parent space plane normal
	real         *destPlaneD   // real to receive the parent space plane D coefficient
	) const
{
	Vector v;

	// rotate the normal into the new space
	rotate(srcNormal, destNormal);

	// compute a point on the plane
	v.mul(srcNormal, -srcPlaneD);

	// rotate the point on the plane into the new space
	rotateTranslate(v, &v);

	// resolve the plane equation in the new space
	*destPlaneD = -destNormal->dot(v);
}

// ------------------------------------------------------------------------
// Rotate and translate an array of source vectors from transform space into parent space, while generating clip
// flag information.
//
// Remarks:
//
//   This routine is used by the 3d graphics pipeline to transform vertices from object space into camera space
//   and generate polygon clipping flags as it goes.
//
//   This routine is highly optimized for the Win95 platform
//
// See also:
//
//   rotateTranslate()

bool Xform::rotateTranslatePerspectiveClip(
	const Vector *src,         // source array of vertices to transform
	Vector       *dest,        // array to receive transformed vertices
	int           cnt,         // number of vertices to transform
	real          nearPlane,   // the distance to the near clip plane
	real          farPlane,    // the distance to the far clip plane
	real          xMult,       // right and left clip plane multipliers
	real          yMult,       // top and bottom clip plane multipliers
	int8         *clipPtr      // array to receive the clip flags
	) const
{
#if PLATFORM == PLATFORM_WIN95

	bool result(false);

	// lint can't see the variable referenced in the inline assembly
	UNREF(src);
	UNREF(dest);
	UNREF(cnt);
	UNREF(nearPlane);
	UNREF(farPlane);
	UNREF(xMult);
	UNREF(yMult);
	UNREF(clipPtr);

	__asm
	{
		mov ebx, clipPtr
		mov ecx, [cnt]
		mov esi, src
		mov edi, dest

		mov dh, 255

		fld DWORD PTR [nearPlane]   // nearPlane
		fld DWORD PTR [farPlane]    // farPlane  nearPlane

	top:

		dec ecx
		js done
	
		// farPlane and nearPlane are going to be on the FPU stack throughout the routine.  I won't bother putting them on
		// every line in this transformation.

		mov eax, this

		                             // s0                   s1                   s2                   s3    s4    s5        

		fld DWORD PTR [esi+0]        // sX
		fld DWORD PTR [esi+4]        // sY                   sX
		fld DWORD PTR [esi+8]        // sZ                   sY                   sX
		fld DWORD PTR [eax+0]        // m00                  sZ                   sY                   sX
		fmul st(0), st(3)            // m00*sX               sZ                   sY                   sX
		fld DWORD PTR [eax+4]        // m01                  m00*sX               sZ                   sY    sX
		fmul st(0), st(3)            // m01*sY               m00*sX               sZ                   sY    sX
		fld DWORD PTR [eax+8]        // m02                  m01*sY               m00*sX               sZ    sY    sX
		fmul st(0), st(3)            // m02*sZ               m01*sY               m00*sX               sZ    sY    sX 
		fxch st(2)                   // m00*sX               m01*sY               m02*sZ               sZ    sY    sX
		faddp st(1), st(0)           // m00*sX+m01*sY        m02*sZ               sZ                   sY    sX
		fld DWORD PTR [eax+16+0]     // m10                  m00*sX+m01*sY        m02*sZ               sZ    sY    sX
		fmul st(0), st(5)            // m10*sX               m00*sX+m01*sY        m02*sZ               sZ    sY    sX
		fxch st(2)                   // m02*sZ               m00*sX+m01*sY        m10*sX               sZ    sY    sX
		faddp st(1), st(0)           // m00*sX+m01*sY+m02*sZ m10*sX               sZ                   sY    sX
		fld DWORD PTR [eax+16+4]     // m11                  m00*sX+m01*sY+m02*sZ m10*sX               sZ    sY    sX
		fmul st(0), st(4)            // m11*sy               m00*sX+m01*sY+m02*sZ m10*sX               sZ    sY    sX
		fxch st(1)                   // m00*sX+m01*sY+m02*sZ m11*sy               m10*sX               sZ    sY    sX
		fadd DWORD PTR [eax+12]      // dx                   m11*sy               m10*sX               sZ    sY    sX
		fstp DWORD PTR [edi+0]       // m11*sy               m10*sX               sZ                   sY    sX
		fld DWORD PTR [eax+16+8]     // m12                  m11*sy               m10*sX               sZ    sY    sX
		fmul st(0), st(3)            // m12*sZ               m11*sy               m10*sX               sZ    sY    sX
		fxch st(2)                   // m10*sX               m11*sy               m12*sZ               sZ    sY    sX
		faddp st(1), st(0)           // m10*sX+m11*sy        m12*sZ               sZ                   sY    sX
		fld DWORD PTR [eax+16+16+0]  // m20                  m10*sX+m11*sy        m12*sZ               sZ    sY    sX
		fmul st(0), st(5)            // m20*sx               m10*sX+m11*sy        m12*sZ               sZ    sY    sX
		fxch st(2)                   // m12*sZ               m10*sX+m11*sy        m20*sx               sZ    sY    sX
		faddp st(1), st(0)           // m10*sX+m11*sy+m12*sZ m20*sx               sZ                   sY    sX
		fld DWORD PTR [eax+16+16+4]  // m21                  m10*sX+m11*sy+m12*sZ m20*sx               sZ    sY    sX
		fmul st(0), st(4)            // m21*sy               m10*sX+m11*sy+m12*sZ m20*sx               sZ    sY    sX
		fxch st(1)                   // m10*sX+m11*sy+m12*sZ m21*sy               m20*sx               sZ    sY    sX
		fadd DWORD PTR [eax+16+12]   // dy                   m21*sy               m20*sx               sZ    sY    sX
		fstp DWORD PTR [edi+4]       // m21*sy               m20*sx               sZ                   sY    sX
		fld DWORD PTR[eax+16+16+8]   // m22                  m21*sy               m20*sx               sZ    sY    sX
		fmul st(0), st(3)            // m22*sz               m21*sy               m20*sx               sZ    sY    sX
		fxch st(2)                   // m20*sx               m21*sy               m22*sz               sZ    sY    sX
		faddp st(1), st(0)           // m20*sx+m21*sy        m22*sz               sZ                   sY    sX
		fxch st(2)                   // sz                   m20*sx+m21*sy        m22*sz               sY    sX
		fstp st(0)                   // m20*sx+m21*sy        m22*sz               sY                   sX
		faddp st(1), st(0)           // m20*sx+m21*sy+m22*sz sY                   sX
		fxch st(2)                   // sx                   sy                   m20*sx+m21*sy+m22*sz
		fstp st(0)                   // sy                   m20*sx+m21*sy+m22*sz
		fstp st(0)                   // m20*sx+m21*sy+m22*sz
		fadd DWORD PTR [eax+16+16+12]// dz
		fst DWORD PTR[edi+8]         // dz

		// compare dz against the near plane
		fcom st(2)                   // dz  farPlane  nearPlane
		xor eax, eax
 		fnstsw ax                    // dz  farPlane  nearPlane
		mov dl, 0
		shr eax, 8
		and eax, 0x01
		or dl, al

		// compare dz against the far plane
		fcom st(1)                   // dz  farPlane  nearPlane
		xor eax, eax
		fnstsw ax                    // dz  farPlane  nearPlane
		shr eax, 7
		and eax, 0x02
		xor eax, 0x02
		or dl, al

		// setup to compare rest of the frustum
		fabs                         // +dz  farPlane  nearPlane
		fld st(0)                    // +dz  +dz  farPlane  nearPlane
		fchs                         // -dz  +dz  farPlane  nearPlane
		fld DWORD PTR [edi+0]        // dx  -dz  +dz  farPlane  nearPlane
		fmul DWORD PTR [xMult]       // dx*xMult  -dz  +dz  farPlane  nearPlane
		fld DWORD PTR [edi+4]        // dy  dx*xMult  -dz  +dz  farPlane  nearPlane
		fmul DWORD PTR [yMult]       // dy*yMult  dx*xMult  -dz  +dz  farPlane  nearPlane

		// compare against the right
		fxch st(1)                   // dx*xMult  dy*yMult  -dz  +dz  farPlane  nearPlane
		fcom st(3)                   // dx*xMult  dy*yMult  -dz  +dz  farPlane  nearPlane
		xor eax, eax
		fnstsw ax                    // dx*xMult  dy*yMult  -dz  +dz  farPlane  nearPlane
		shr eax, 5
		and eax, 0x08
		fcomp st(2)                  // dy*yMult  -dz  +dz  farPlane  nearPlane
		xor eax, 0x08
		or dl, al

		// compare against the left
		xor eax, eax
		fnstsw ax                    // dy*yMult  -dz  +dz  farPlane  nearPlane
		shr eax, 3
		and eax, 0x20
		fcom st(2)                   // dy*yMult  -dz  +dz  farPlane  nearPlane
		or dl, al

		// compare against top
		xor eax, eax
		fnstsw ax                    // dy*yMult  -dz  +dz  farPlane  nearPlane
		shr eax, 6
		and eax, 0x04
		fcomp st(1)                  // -dz  +dz  farPlane  nearPlane
		xor eax, 0x04
		or dl, al

		// compare against bottom
		xor eax, eax
		fnstsw ax                    // -dz  +dz  farPlane  nearPlane
		shr eax, 4
		and eax, 0x10
		or dl, al

		// pop off -dz and +dz
		fstp st(0)                   // +dz  farPlane  nearPlane
		fstp st(0)                   // farPlane  nearPlane
				
		// update the clip flags
		mov byte ptr [ebx], dl
		add ebx, 1

		// keep track of the and'ed clip flags
		and dh, dl

		// advance src & dest
		add esi, 12
		add edi, 12

		jmp top

	done:
		fstp st(0)
		fstp st(0)

		test dh, dh
		mov dh, 0
		jnz over
		mov dh, 1

	over:
		mov byte ptr [result], dh
	}

	return result;

#else

	int8 clipAnd, clip;
	real x, y, z, dx, dy, dz;
	
	for (clipAnd = CLIP_ALL; cnt--; src++, dest++, clipPtr++)
	{
		x = src->x;
		y = src->y;
		z = src->z;
		
		// rotate and translate the vector
		dest->x = dx = matrix[0][0]*x + matrix[0][1]*y + matrix[0][2]*z + matrix[0][3];
		dest->y = dy = matrix[1][0]*x + matrix[1][1]*y + matrix[1][2]*z + matrix[1][3];
		dest->z = dz = matrix[2][0]*x + matrix[2][1]*y + matrix[2][2]*z + matrix[2][3];

		// compute the clipping flags
		clip = 0;
		if (dz < nearPlane)
			clip |= CLIP_NEAR;
		if (dz > farPlane)
			clip |= CLIP_FAR;

		// get positive z
		dx *= xMult;
		dy *= yMult;
		dz  = abs(dz);

		if (dy >  dz)
			clip |= CLIP_ABOVE;
		if (dx >  dz)
			clip |= CLIP_RIGHT;
		if (dy < -dz)
			clip |= CLIP_BELOW;
		if (dx < -dz) 
			clip |= CLIP_LEFT;

		*clipPtr = clip;

		// and the flags as we go
		clipAnd &= clip;
	}

	return (clipAnd == 0);

#endif
}

// ----------------------------------------------------------------------
// Orient this transform to the specified vectors
//
// Remarks:
//
//   This routine will orient this transformation according to the three vectors passed to it.  Only one of the
//   three vector pointers may be null, in which case this routine will calculate it.
//
//   If the vectors passed in are not unit vectors, or may not be orthogonal with respect to each other, the
//   needsReorthongalize value must be true, otherwise it may be false for efficiency.

void Xform::rotateToFrame(
	const Vector *unitI,       // vector pointing down the X axis
	const Vector *unitJ,       // vector pointing down the Y axis
	const Vector *unitK,       // vector pointing down the Z axis
	bool needsReorthogonalize  // whether to automatically reorthogonalize this transform or not
	)
{
	Vector tmp;

	if (!unitI && unitJ && unitK)
	{
		tmp.cross(*unitJ, *unitK);
		unitI = &tmp;
	}

	if (!unitJ && unitI && unitK)
	{
		tmp.cross(*unitK, *unitI);
		unitJ = &tmp;
	}

	if (!unitK && unitI && unitJ)
	{
		tmp.cross(*unitI, *unitJ);
		unitK = &tmp;
	}

	DEBUG_FATAL((!unitI || !unitJ || !unitK), ("Xform::rotateToFrame need at least two vectors (%p %p %p)", unitI, unitJ, unitK));

	// x coords
	matrix[0][0] = unitI->x;
	matrix[1][0] = unitI->y;
	matrix[2][0] = unitI->z;

	// y coords
	matrix[0][1] = unitJ->x;
	matrix[1][1] = unitJ->y;
	matrix[2][1] = unitJ->z;

	// z coords
	matrix[0][2] = unitK->x;
	matrix[1][2] = unitK->y;
	matrix[2][2] = unitK->z;

	if (needsReorthogonalize)
		reorthogonalize();
}

// ----------------------------------------------------------------------
// Break down matrix in yaw, pitch, roll order
//
// Remarks:
//
//   This routine is slow.  It requires at least one asin(), and at least one atan2() (but likely two).

void Xform::getOrientationYPR(
	real *pYawRadians,    // pointer to the real to receive the yaw, may be NULL
	real *pPitchRadians,  // pointer to the real to receive the pitch, may be NULL
	real *pRollRadians    // pointer to the real to receive the roll, may be NULL
	) const
{
	real   yawRadians, pitchRadians, rollRadians;

#if 0

	yawRadians = asin(matrix[0][2]);

	// is the abs(yaw) PI / 2?  if so avoid dividing by 0
	if (abs(yawRadians) != Pi_1_2)
	{
		pitchRadians = -atan2(matrix[1][2], matrix[2][2]);
		rollRadians  = -atan2(matrix[0][1], matrix[0][0]);
	}
	else
	{
		pitchRadians = -atan2(matrix[1][0], matrix[1][1]);
		rollRadians  = CONST_REAL(0);
	}

#else

	real cy = sqrt (matrix[1][1] * matrix[1][1] + matrix[0][1] * matrix[0][1]);
	if (cy > 16 * FLT_EPSILON) 
	{
		yawRadians   = -atan2 ( matrix[2][0], matrix[2][2]);
		pitchRadians = -atan2 (-matrix[2][1], cy);
		rollRadians  = -atan2 ( matrix[0][1], matrix[1][1]);
	}
	else 
	{
		yawRadians   = -atan2 (-matrix[0][2], matrix[0][0]);
		pitchRadians = -atan2 (-matrix[2][1], cy);
		rollRadians  = 0;
	}

#endif

	if (pYawRadians)
		*pYawRadians   = yawRadians;
	if (pPitchRadians)
		*pPitchRadians = pitchRadians;
	if (pRollRadians)
		*pRollRadians  = rollRadians;
 }

// ----------------------------------------------------------------------
// Reorthogonalize this transform
//
// Remarks:
//
//   Applying a series of rotations will introduce numerical error that causes the three vectors in the upper 3x3
//   of the transform to become non-orthogonal and non-unit vectors.  Calling this routine form a transform will
//   reduce the numerical error by reconstructing the transform based on its current data.

void Xform::reorthogonalize(void)
{
	Vector i, j, k;

	// we're about to compute i from scratch, so don't bother fetching it

	j.x = matrix[0][1];
	j.y = matrix[1][1];
	j.z = matrix[2][1];

	k.x = matrix[0][2];
	k.y = matrix[1][2];
	k.z = matrix[2][2];

	// normalize the vectors to keep us from going wonkers
	if (!j.normalize())
		DEBUG_FATAL(true, ("Xform::reorthogonalize j failed"));
	if (!k.normalize())
		DEBUG_FATAL(true, ("Xform::reorthogonalize k failed"));

	// regenerate i from the normalized j and k vectors
	i.cross(j, k);

	// regenerate j from k and the new i
	j.cross(k, i);

	// put our results back into the matrix
	matrix[0][0] = i.x;
	matrix[1][0] = i.y;
	matrix[2][0] = i.z;

	matrix[0][1] = j.x;
	matrix[1][1] = j.y;
	matrix[2][1] = j.z;

	matrix[0][2] = k.x;
	matrix[1][2] = k.y;
	matrix[2][2] = k.z;
}

// ----------------------------------------------------------------------

void Xform::writeSave(Save *save)
{
	int    i, j, k;

	for (k = 0, i = 0; i < 3; i++)
		for (j = 0; j < 4; j++, k++)
			save->save[k] = matrix[i][j];
}

// ----------------------------------------------------------------------

void Xform::readSave(const Save &save)
{
	int    i, j, k;

	for (k = 0, i = 0; i < 3; i++)
		for (j = 0; j < 4; j++, k++)
			matrix[i][j] = save.save[k];
}

// ======================================================================
